/*
 *  $Id: psf.c 23556 2021-04-21 08:00:51Z yeti-dn $
 *  Copyright (C) 2017-2019 David Necas (Yeti), Petr Klapetek.
 *  E-mail: yeti@gwyddion.net, klapetek@gwyddion.net.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <fftw3.h>
#include <gtk/gtk.h>
#include <libgwyddion/gwymacros.h>
#include <libgwyddion/gwymath.h>
#include <libgwyddion/gwythreads.h>
#include <libprocess/gwyprocesstypes.h>
#include <libprocess/arithmetic.h>
#include <libprocess/grains.h>
#include <libprocess/inttrans.h>
#include <libprocess/filters.h>
#include <libprocess/stats.h>
#include <libprocess/simplefft.h>
#include <libgwydgets/gwydgetutils.h>
#include <libgwydgets/gwycombobox.h>
#include <libgwydgets/gwycheckboxes.h>
#include <libgwydgets/gwystock.h>
#include <libgwymodule/gwymodule-process.h>
#include <libgwymodule/gwymodule-file.h>
#include <app/gwymoduleutils.h>
#include <app/gwyapp.h>
#include "libgwyddion/gwyomp.h"
#include "mfmops.h"
#include "preview.h"

#define PSF_RUN_MODES GWY_RUN_INTERACTIVE

#define field_convolve_default(field, kernel) \
    gwy_data_field_area_ext_convolve((field), \
                                     0, 0, \
                                     gwy_data_field_get_xres(field), \
                                     gwy_data_field_get_yres(field), \
                                     (field), (kernel), \
                                     GWY_EXTERIOR_BORDER_EXTEND, 0.0, TRUE)

#define gwycreal(x) ((x)[0])
#define gwycimag(x) ((x)[1])

/* Do not require FFTW 3.3 just for a couple of trivial macros. */
#define gwy_fftw_new_real(n) \
    (gdouble*)fftw_malloc((n)*sizeof(gdouble))
#define gwy_fftw_new_complex(n) \
    (fftw_complex*)fftw_malloc((n)*sizeof(fftw_complex))

typedef enum {
    PSF_METHOD_REGULARISED   = 0,
    PSF_METHOD_LEAST_SQUARES = 1,
    PSF_METHOD_PSEUDO_WIENER = 2,
    PSF_NMETHODS
} PSFMethodType;

typedef enum {
    PSF_DISPLAY_DATA       = 0,
    PSF_DISPLAY_PSF        = 1,
    PSF_DISPLAY_CONVOLVED  = 2,
    PSF_DISPLAY_DIFFERENCE = 3,
    PSF_NDISPLAYS
} PSFDisplayType;

typedef enum {
    PSF_OUTPUT_PSF       = (1 << 0),
    PSF_OUTPUT_CONVOLVED = (1 << 1),
    PSF_OUTPUT_DIFFERECE = (1 << 2),
    PSF_OUTPUTS_MASK     = (1 << 3) - 1,
} PSFOutputType;

typedef struct {
    PSFMethodType method;
    gdouble sigma;                /* Log10 */
    GwyWindowingType windowing;
    GwyAppDataId op1;
    GwyAppDataId op2;
    gint txres;
    gint tyres;
    gint border;
    PSFDisplayType display;
    PSFOutputType output_type;
    gboolean as_integral;
} PSFArgs;

static const gchar *guivalues[] = {
    "width", "height", "l2norm", "residuum",
};

typedef struct {
    PSFArgs *args;
    GtkWidget *method;
    GtkObject *sigma;
    GtkWidget *chooser_op2;
    GtkWidget *display;
    GtkWidget *windowing;
    GtkObject *txres;
    GtkObject *tyres;
    GtkObject *border;
    GtkWidget *guess_tres;
    GtkWidget *full_tres;
    GSList *output_type;
    GtkWidget *as_integral;
    GtkWidget *tf_size_header;
    GwyDataField *result;
    GtkWidget *view;
    GwyContainer *mydata;
    GtkWidget *dialog;
    gboolean in_update;
    GwyResults *results;
    GtkWidget *guivalues[G_N_ELEMENTS(guivalues)];
} PSFControls;

static gboolean module_register          (void);
static void     psf                      (GwyContainer *data,
                                          GwyRunType run);
static gboolean psf_dialog               (PSFArgs *args,
                                          GwyContainer *data);
static void     psf_load_args            (GwyContainer *container,
                                          PSFArgs *args);
static void     psf_save_args            (GwyContainer *container,
                                          PSFArgs *args);
static void     psf_sanitize_args        (PSFArgs *args);
static void     method_changed           (GtkComboBox *combo,
                                          PSFControls *controls);
static void     sigma_changed            (GtkAdjustment *adj,
                                          PSFControls *controls);
static void     windowing_changed        (GtkComboBox *combo,
                                          PSFControls *controls);
static void     txres_changed            (GtkAdjustment *adj,
                                          PSFControls *controls);
static void     tyres_changed            (GtkAdjustment *adj,
                                          PSFControls *controls);
static void     border_changed           (GtkAdjustment *adj,
                                          PSFControls *controls);
static void     guess_tres_clicked       (GtkButton *button,
                                          PSFControls *controls);
static void     full_tres_clicked        (GtkButton *button,
                                          PSFControls *controls);
static void     psf_data_changed         (GwyDataChooser *chooser,
                                          PSFControls *controls);
static gboolean psf_data_filter          (GwyContainer *data,
                                          gint id,
                                          gpointer user_data);
static void     preview                  (PSFControls *controls,
                                          PSFArgs *args);
static void     display_changed          (GtkComboBox *combo,
                                          PSFControls *controls);
static void     as_integral_changed      (GtkToggleButton *toggle,
                                          PSFControls *controls);
static void     output_type_changed      (GtkToggleButton *toggle,
                                          PSFControls *controls);
static void     update_sensitivity       (PSFControls *controls);
static void     update_tres_for_method   (PSFControls *controls);
static void     prepare_field            (GwyDataField *field,
                                          GwyDataField *wfield,
                                          GwyWindowingType window);
static gboolean method_is_full_sized     (PSFMethodType method);
static void     estimate_tf_region       (GwyDataField *wmeas,
                                          GwyDataField *wideal,
                                          GwyDataField *psf,
                                          gint *col,
                                          gint *row,
                                          gint *width,
                                          gint *height);
static void     symmetrise_tf_region     (gint pos,
                                          gint len,
                                          gint res,
                                          gint *tres);
static void     calculate_tf             (GwyDataField *measured,
                                          GwyDataField *ideal,
                                          GwyDataField *psf,
                                          PSFArgs *args);
static void     psf_deconvolve_wiener    (GwyDataField *dfield,
                                          GwyDataField *operand,
                                          GwyDataField *out,
                                          gdouble sigma);
static void     adjust_tf_to_non_integral(GwyDataField *psf);
static gdouble  measure_tf_width         (GwyDataField *psf);
static void     estimate_sigma           (PSFArgs *args);
static gdouble  find_regularization_sigma(GwyDataField *dfield,
                                          GwyDataField *ideal,
                                          const PSFArgs *args);
static gint     create_output_field      (GwyDataField *dfield,
                                          GwyAppDataId dataid,
                                          const gchar *name);
static gboolean clamp_psf_size           (GwyDataField *dfield,
                                          PSFArgs *args);

static GwyAppDataId op2_id = GWY_APP_DATA_ID_NONE;

static const PSFArgs psf_defaults = {
    PSF_METHOD_REGULARISED, 1.0, GWY_WINDOWING_WELCH,
    GWY_APP_DATA_ID_NONE, GWY_APP_DATA_ID_NONE,
    51, 51, 3,
    PSF_DISPLAY_PSF,
    PSF_OUTPUT_PSF,
    TRUE,
};

static const GwyEnum output_types[] = {
    { N_("Transfer function"), PSF_OUTPUT_PSF,       },
    { N_("Convolved"),         PSF_OUTPUT_CONVOLVED, },
    { N_("Difference"),        PSF_OUTPUT_DIFFERECE, },
};

enum {
    OUTPUT_NTYPES = G_N_ELEMENTS(output_types)
};

static GwyModuleInfo module_info = {
    GWY_MODULE_ABI_VERSION,
    &module_register,
    N_("Transfer function estimation"),
    "Petr Klapetek <klapetek@gwyddion.net>, Yeti <yeti@gwyddion.net>",
    "3.2",
    "David Nečas (Yeti) & Petr Klapetek",
    "2017",
};

GWY_MODULE_QUERY2(module_info, psf)

static gboolean
module_register(void)
{
    gwy_process_func_register("psf",
                              (GwyProcessFunc)&psf,
                              N_("/_Statistics/_Transfer Function Guess..."),
                              NULL,
                              PSF_RUN_MODES,
                              GWY_MENU_FLAG_DATA,
                              N_("Estimate transfer function "
                                 "from known data and ideal image"));

    return TRUE;
}

static void
psf(GwyContainer *data, GwyRunType run)
{
    GwyDataField *dfield1, *dfield2, *psf, *convolved = NULL, *difference;
    PSFArgs args;
    gboolean ok;
    GQuark quark;
    GwyContainer *data1, *data2;
    gint newid = -1;

    g_return_if_fail(run & PSF_RUN_MODES);

    psf_load_args(gwy_app_settings_get(), &args);

    gwy_app_data_browser_get_current(GWY_APP_DATA_FIELD_ID, &args.op1.id,
                                     GWY_APP_CONTAINER_ID, &args.op1.datano,
                                     0);

    data1 = gwy_app_data_browser_get(args.op1.datano);
    quark = gwy_app_get_data_key_for_id(args.op1.id);
    dfield1 = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));

    if (!clamp_psf_size(dfield1, &args)) {
        if (run == GWY_RUN_INTERACTIVE) {
            GtkWidget *dialog;

            dialog = gtk_message_dialog_new
                           (gwy_app_find_window_for_channel(data, args.op1.id),
                            GTK_DIALOG_DESTROY_WITH_PARENT,
                            GTK_MESSAGE_ERROR,
                            GTK_BUTTONS_OK,
                            _("Image is too small."));
            gtk_dialog_run(GTK_DIALOG(dialog));
            gtk_widget_destroy(dialog);
        }
        return;
    }

    ok = psf_dialog(&args, data);
    psf_save_args(gwy_app_settings_get(), &args);
    if (!ok)
        return;

    data2 = gwy_app_data_browser_get(args.op2.datano);
    quark = gwy_app_get_data_key_for_id(args.op2.id);
    dfield2 = GWY_DATA_FIELD(gwy_container_get_object(data2, quark));

    psf = gwy_data_field_new_alike(dfield1, TRUE);
    calculate_tf(dfield1, dfield2, psf, &args);
    if (args.output_type & (PSF_OUTPUT_CONVOLVED | PSF_OUTPUT_DIFFERECE)) {
        convolved = gwy_data_field_duplicate(dfield2);
        gwy_data_field_add(convolved, -gwy_data_field_get_avg(convolved));
        field_convolve_default(convolved, psf);
        gwy_data_field_add(convolved, gwy_data_field_get_avg(dfield1));
    }

    if (args.output_type & PSF_OUTPUT_PSF)
        newid = create_output_field(psf, args.op1, _("Transfer function"));
    if (args.output_type & PSF_OUTPUT_CONVOLVED)
        create_output_field(convolved, args.op1, _("Convolved"));
    if (args.output_type & PSF_OUTPUT_DIFFERECE) {
        difference = gwy_data_field_new_alike(dfield1, FALSE);
        gwy_data_field_subtract_fields(difference, convolved, dfield1);
        create_output_field(difference, args.op1, _("Difference"));
        g_object_unref(difference);
    }

    GWY_OBJECT_UNREF(convolved);
    g_object_unref(psf);

    /* Change the normalisation to the discrete (i.e. wrong) one after all
     * calculations are done. */
    if (!args.as_integral && newid != -1) {
        data = gwy_app_data_browser_get(args.op1.datano);
        quark = gwy_app_get_data_key_for_id(newid);
        psf = gwy_container_get_object(data, quark);
        adjust_tf_to_non_integral(psf);
    }
}

static gboolean
psf_dialog(PSFArgs *args, GwyContainer *data)
{
    static const GwyEnum psf_methods[] = {
        { N_("Regularized filter"), PSF_METHOD_REGULARISED,   },
        { N_("Least squares"),      PSF_METHOD_LEAST_SQUARES, },
        { N_("Wiener filter"),      PSF_METHOD_PSEUDO_WIENER, },
    };
    static const GwyEnum psf_displays[] = {
        { N_("Data"),              PSF_DISPLAY_DATA,       },
        { N_("Transfer function"), PSF_DISPLAY_PSF,        },
        { N_("Convolved"),         PSF_DISPLAY_CONVOLVED,  },
        { N_("Difference"),        PSF_DISPLAY_DIFFERENCE, },
    };

    GtkWidget *dialog, *table, *hbox, *label, *align, *notebook, *hbox2;
    GwyDataChooser *chooser;
    PSFControls controls;
    gint response, row, xres, yres;
    GQuark quark;
    GwyContainer *data1;
    GwyResults *results;
    GwyDataField *dfield;
    GString *str;
    guint i;

    gwy_clear(&controls, 1);
    controls.args = args;
    controls.in_update = TRUE;

    controls.results = results = gwy_results_new();
    gwy_results_add_value_x(results, "width", N_("TF width"));
    gwy_results_add_value_z(results, "height", N_("TF height"));
    gwy_results_add_value(results, "l2norm", N_("TF norm"), "power-u", 1, NULL);
    gwy_results_add_value(results, "residuum", N_("Difference norm"),
                          "power-v", 1,
                          NULL);

    dialog = gtk_dialog_new_with_buttons(_("Estimate Transfer Function"),
                                         NULL, 0,
                                         _("_Fit Sigma"), RESPONSE_ESTIMATE,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK,
                                         NULL);
    gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
    controls.dialog = dialog;

    hbox = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), GTK_WIDGET(hbox),
                       FALSE, FALSE, 4);

    data1 = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    dfield = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));
    controls.result = gwy_data_field_new_alike(dfield, TRUE);
    xres = gwy_data_field_get_xres(dfield);
    yres = gwy_data_field_get_yres(dfield);

    controls.mydata = gwy_container_new();
    gwy_container_set_object_by_name(controls.mydata, "/0/data",
                                     controls.result);
    gwy_app_sync_data_items(data, controls.mydata, args->op1.id, 0, FALSE,
                            GWY_DATA_ITEM_GRADIENT,
                            GWY_DATA_ITEM_REAL_SQUARE,
                            0);
    controls.view = gwy_create_preview(controls.mydata, 0, PREVIEW_SIZE, FALSE);
    align = gtk_alignment_new(0.5, 0.0, 0, 0);
    gtk_container_add(GTK_CONTAINER(align), controls.view);
    gtk_box_pack_start(GTK_BOX(hbox), align, FALSE, FALSE, 4);

    notebook = gtk_notebook_new();
    gtk_box_pack_start(GTK_BOX(hbox), notebook, TRUE, TRUE, 4);

    table = gtk_table_new(10 + G_N_ELEMENTS(guivalues), 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 6);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), table,
                             gtk_label_new(_("Parameters")));
    row = 0;

    controls.chooser_op2 = gwy_data_chooser_new_channels();
    chooser = GWY_DATA_CHOOSER(controls.chooser_op2);
    gwy_data_chooser_set_active_id(chooser, &args->op2);
    gwy_data_chooser_set_filter(chooser, psf_data_filter, &args->op1, NULL);
    g_signal_connect(chooser, "changed",
                     G_CALLBACK(psf_data_changed), &controls);
    gwy_table_attach_adjbar(table, row, _("_Ideal response:"), NULL,
                            GTK_OBJECT(chooser), GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    controls.method
        = gwy_enum_combo_box_new(psf_methods, G_N_ELEMENTS(psf_methods),
                                 G_CALLBACK(method_changed), &controls,
                                 args->method, TRUE);
    gwy_table_attach_adjbar(GTK_WIDGET(table), row, _("_Method:"), NULL,
                            GTK_OBJECT(controls.method),
                            GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;

    controls.sigma = gtk_adjustment_new(args->sigma, -8.0, 3.0, 0.001, 1.0, 0);
    gwy_table_attach_adjbar(table, row, _("_Sigma:"), "log<sub>10</sub>",
                            controls.sigma, GWY_HSCALE_LINEAR);
    g_signal_connect(controls.sigma, "value-changed",
                     G_CALLBACK(sigma_changed), &controls);
    row++;

    controls.windowing
        = gwy_enum_combo_box_new(gwy_windowing_type_get_enum(), -1,
                                 G_CALLBACK(windowing_changed), &controls,
                                 args->windowing, TRUE);
    gwy_table_attach_adjbar(GTK_WIDGET(table), row, _("_Windowing type:"), NULL,
                            GTK_OBJECT(controls.windowing),
                            GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    controls.tf_size_header = gwy_label_new_header(_("Transfer Function Size"));
    gtk_table_attach(GTK_TABLE(table), controls.tf_size_header,
                     0, 3, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    controls.txres = gtk_adjustment_new(args->txres, 3, xres, 2, 10, 0);
    gwy_table_attach_adjbar(table, row, _("_Horizontal size:"), _("px"),
                            controls.txres, GWY_HSCALE_SQRT | GWY_HSCALE_SNAP);
    g_signal_connect(controls.txres, "value-changed",
                     G_CALLBACK(txres_changed), &controls);
    row++;

    controls.tyres = gtk_adjustment_new(args->tyres, 3, yres, 2, 10, 0);
    gwy_table_attach_adjbar(table, row, _("_Vertical size:"), _("px"),
                            controls.tyres, GWY_HSCALE_SQRT | GWY_HSCALE_SNAP);
    g_signal_connect(controls.tyres, "value-changed",
                     G_CALLBACK(tyres_changed), &controls);
    row++;

    controls.border = gtk_adjustment_new(args->border,
                                         0, MIN(xres, yres)/8, 1, 5, 0);
    gwy_table_attach_adjbar(table, row, _("_Border:"), _("px"),
                            controls.border, GWY_HSCALE_SQRT | GWY_HSCALE_SNAP);
    g_signal_connect(controls.border, "value-changed",
                     G_CALLBACK(border_changed), &controls);
    row++;

    hbox2 = gtk_hbox_new(FALSE, 2);
    gtk_table_attach(GTK_TABLE(table), hbox2,
                     0, 2, row, row+1, GTK_FILL, 0, 0, 0);
    controls.guess_tres = gtk_button_new_with_mnemonic("_Estimate Size");
    gtk_box_pack_end(GTK_BOX(hbox2), controls.guess_tres, FALSE, FALSE, 0);
    g_signal_connect(controls.guess_tres, "clicked",
                     G_CALLBACK(guess_tres_clicked), &controls);
    controls.full_tres = gtk_button_new_with_mnemonic("_Full Size");
    gtk_box_pack_end(GTK_BOX(hbox2), controls.full_tres, FALSE, FALSE, 0);
    g_signal_connect(controls.full_tres, "clicked",
                     G_CALLBACK(full_tres_clicked), &controls);
    row++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    label = gwy_label_new_header(_("Preview Options"));
    gtk_table_attach(GTK_TABLE(table), label,
                     0, 3, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    controls.display
        = gwy_enum_combo_box_new(psf_displays, G_N_ELEMENTS(psf_displays),
                                 G_CALLBACK(display_changed),
                                 &controls, args->display, TRUE);
    gwy_table_attach_adjbar(table, row, _("_Display:"), NULL,
                            GTK_OBJECT(controls.display),
                            GWY_HSCALE_WIDGET_NO_EXPAND);
    row++;

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    label = gwy_label_new_header(_("Result"));
    gtk_table_attach(GTK_TABLE(table), label,
                     0, 3, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    str = g_string_new(NULL);
    for (i = 0; i < G_N_ELEMENTS(guivalues); i++) {
        g_string_assign(str, gwy_results_get_label_with_symbol(results,
                                                               guivalues[i]));
        g_string_append_c(str, ':');
        label = gtk_label_new(str->str);
        gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_table_attach(GTK_TABLE(table), label, 0, 1, row, row+1,
                         GTK_FILL, 0, 0, 0);

        label = controls.guivalues[i] = gtk_label_new(NULL);
        gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
        gtk_label_set_selectable(GTK_LABEL(label), TRUE);
        gtk_table_attach(GTK_TABLE(table), label, 1, 2, row, row+1,
                         GTK_EXPAND | GTK_FILL, 0, 0, 0);
        row++;
    }
    g_string_free(str, TRUE);

    table = gtk_table_new(3 + OUTPUT_NTYPES, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 2);
    gtk_table_set_col_spacings(GTK_TABLE(table), 6);
    gtk_container_set_border_width(GTK_CONTAINER(table), 4);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), table,
                             gtk_label_new(_("Output Options")));
    row = 0;

    label = gtk_label_new(_("Output type:"));
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 0, 2, row, row+1,
                     GTK_FILL, 0, 0, 0);
    row++;

    controls.output_type
        = gwy_check_boxes_create(output_types, OUTPUT_NTYPES,
                                 G_CALLBACK(output_type_changed), &controls,
                                 args->output_type);
    row = gwy_check_boxes_attach_to_table(controls.output_type,
                                          GTK_TABLE(table), 2, row);

    gtk_table_set_row_spacing(GTK_TABLE(table), row-1, 8);
    controls.as_integral
        = gtk_check_button_new_with_mnemonic(_("Normalize as _integral"));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(controls.as_integral),
                                 args->as_integral);
    g_signal_connect(controls.as_integral, "toggled",
                     G_CALLBACK(as_integral_changed), &controls);
    gtk_table_attach(GTK_TABLE(table), controls.as_integral,
                     0, 2, row, row+1, GTK_FILL, 0, 0, 0);
    row++;

    controls.in_update = FALSE;
    update_tres_for_method(&controls);
    psf_data_changed(chooser, &controls);

    gtk_widget_show_all(dialog);
    do {
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response) {
            case GTK_RESPONSE_CANCEL:
            case GTK_RESPONSE_DELETE_EVENT:
            gtk_widget_destroy(dialog);
            g_object_unref(controls.mydata);
            g_object_unref(controls.results);
            case GTK_RESPONSE_NONE:
            return FALSE;
            break;

            case GTK_RESPONSE_OK:
            break;

            case RESPONSE_ESTIMATE:
            gwy_app_wait_cursor_start(GTK_WINDOW(dialog));
            estimate_sigma(args);
            gwy_app_wait_cursor_finish(GTK_WINDOW(dialog));
            gtk_adjustment_set_value(GTK_ADJUSTMENT(controls.sigma),
                                     args->sigma);
            preview(&controls, args);
            break;

            default:
            g_assert_not_reached();
            break;
        }
    } while (response != GTK_RESPONSE_OK);

    gtk_widget_destroy(dialog);
    g_object_unref(controls.mydata);
    g_object_unref(controls.results);

    return TRUE;
}

static void
display_changed(GtkComboBox *combo, PSFControls *controls)
{
    controls->args->display = gwy_enum_combo_box_get_active(combo);
    preview(controls, controls->args);
}

static void
as_integral_changed(GtkToggleButton *toggle, PSFControls *controls)
{
    controls->args->as_integral = gtk_toggle_button_get_active(toggle);
    /* PSF height changes, unfortunately. */
    preview(controls, controls->args);
}

static void
method_changed(GtkComboBox *combo, PSFControls *controls)
{
    controls->args->method = gwy_enum_combo_box_get_active(combo);
    update_sensitivity(controls);
    update_tres_for_method(controls);
    preview(controls, controls->args);
}

static void
sigma_changed(GtkAdjustment *adj, PSFControls *controls)
{
    controls->args->sigma = gtk_adjustment_get_value(adj);
    preview(controls, controls->args);
}

static void
windowing_changed(GtkComboBox *combo, PSFControls *controls)
{
    controls->args->windowing = gwy_enum_combo_box_get_active(combo);
    preview(controls, controls->args);
}

static void
txres_changed(GtkAdjustment *adj, PSFControls *controls)
{
    controls->args->txres = gwy_adjustment_get_int(adj);
    if (!controls->in_update)
        preview(controls, controls->args);
}

static void
tyres_changed(GtkAdjustment *adj, PSFControls *controls)
{
    controls->args->tyres = gwy_adjustment_get_int(adj);
    if (!controls->in_update)
        preview(controls, controls->args);
}

static void
border_changed(GtkAdjustment *adj, PSFControls *controls)
{
    controls->args->border = gwy_adjustment_get_int(adj);
    preview(controls, controls->args);
}

static void
guess_tres_clicked(G_GNUC_UNUSED GtkButton *button,
                   PSFControls *controls)
{
    PSFArgs *args = controls->args;
    GwyContainer *data1, *data2;
    GQuark quark;
    GwyDataField *measured, *ideal, *wideal, *wmeas, *psf;
    gint col, row, width, height;

    data1 = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    measured = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));

    data2 = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    ideal = GWY_DATA_FIELD(gwy_container_get_object(data2, quark));

    wmeas = gwy_data_field_new_alike(measured, FALSE);
    wideal = gwy_data_field_new_alike(ideal, FALSE);
    prepare_field(measured, wmeas, args->windowing);
    prepare_field(ideal, wideal, args->windowing);

    psf = gwy_data_field_new_alike(measured, TRUE);
    estimate_tf_region(wmeas, wideal, psf, &col, &row, &width, &height);
    g_object_unref(psf);
    g_object_unref(wideal);
    g_object_unref(wmeas);
    symmetrise_tf_region(col, width, gwy_data_field_get_xres(ideal),
                         &args->txres);
    symmetrise_tf_region(row, height, gwy_data_field_get_yres(ideal),
                         &args->tyres);
    args->border = GWY_ROUND(0.5*log(fmax(args->txres, args->tyres)) + 0.5);

    controls->in_update = TRUE;
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->border), args->border);
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->txres), args->txres);
    controls->in_update = FALSE;
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->tyres), args->tyres);
}

static void
full_tres_clicked(G_GNUC_UNUSED GtkButton *button,
                  PSFControls *controls)
{
    PSFArgs *args = controls->args;
    GwyContainer *data1;
    GQuark quark;
    GwyDataField *measured;

    data1 = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    measured = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));

    args->txres = gwy_data_field_get_xres(measured);
    args->tyres = gwy_data_field_get_yres(measured);

    controls->in_update = TRUE;
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->txres), args->txres);
    controls->in_update = FALSE;
    gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->tyres), args->tyres);
}

static void
output_type_changed(G_GNUC_UNUSED GtkToggleButton *toggle,
                    PSFControls *controls)
{
    PSFArgs *args = controls->args;

    args->output_type = gwy_check_boxes_get_selected(controls->output_type);
    update_sensitivity(controls);
}

static void
update_sensitivity(PSFControls *controls)
{
    PSFArgs *args = controls->args;
    gboolean have_ideal = args->op2.datano;
    gboolean out_is_psf = (args->output_type & PSF_OUTPUT_PSF);
    gboolean method_is_lsq = (args->method == PSF_METHOD_LEAST_SQUARES);
    gboolean any_output = args->output_type;

    gtk_dialog_set_response_sensitive(GTK_DIALOG(controls->dialog),
                                      GTK_RESPONSE_OK,
                                      have_ideal && any_output);
    gtk_dialog_set_response_sensitive(GTK_DIALOG(controls->dialog),
                                      RESPONSE_ESTIMATE, have_ideal);
    gtk_widget_set_sensitive(controls->as_integral, out_is_psf);
    gtk_widget_set_sensitive(controls->full_tres, !method_is_lsq);
    gwy_table_hscale_set_sensitive(controls->border, method_is_lsq);
}

static void
update_tres_for_method(PSFControls *controls)
{
    PSFArgs *args = controls->args;
    GwyContainer *data1;
    GQuark quark;
    GwyDataField *measured;
    gint xres, yres, wantxupper, wantyupper;
    gdouble xupper, yupper;

    data1 = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    measured = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));

    xres = gwy_data_field_get_xres(measured);
    yres = gwy_data_field_get_yres(measured);

    g_object_get(controls->txres, "upper", &xupper, NULL);
    g_object_get(controls->tyres, "upper", &yupper, NULL);

    controls->in_update = TRUE;
    if (method_is_full_sized(args->method)) {
        wantxupper = xres;
        wantyupper = yres;
    }
    else {
        wantxupper = (xres/3) | 1;
        wantyupper = (yres/3) | 1;
    }
    if (args->txres > wantxupper)
        gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->txres), wantxupper);
    if (xupper > wantxupper)
        g_object_set(controls->txres, "upper", (gdouble)wantxupper, NULL);
    if (args->tyres > wantyupper)
        gtk_adjustment_set_value(GTK_ADJUSTMENT(controls->tyres), wantyupper);
    if (yupper > wantyupper)
        g_object_set(controls->tyres, "upper", (gdouble)wantyupper, NULL);
    controls->in_update = FALSE;
}

static void
estimate_sigma(PSFArgs *args)
{
    GwyDataField *measured, *ideal;
    GQuark quark;
    GwyContainer *data1, *data2;

    data1 = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    measured = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));

    data2 = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    ideal = GWY_DATA_FIELD(gwy_container_get_object(data2, quark));

    args->sigma = log(find_regularization_sigma(measured, ideal, args))/G_LN10;
}

static gint
create_output_field(GwyDataField *dfield,
                    GwyAppDataId dataid,
                    const gchar *name)
{
    GwyContainer *data;
    gint newid;

    data = gwy_app_data_browser_get(dataid.datano);
    newid = gwy_app_data_browser_add_data_field(dfield, data, TRUE);

    gwy_app_sync_data_items(data, data, dataid.id, newid, FALSE,
                            GWY_DATA_ITEM_GRADIENT,
                            GWY_DATA_ITEM_MASK_COLOR,
                            0);
    gwy_app_set_data_field_title(data, newid, name);
    gwy_app_channel_log_add_proc(data, dataid.id, newid);

    return newid;
}

static gdouble
calculate_l2_norm(GwyDataField *dfield, gboolean as_integral,
                  GwySIUnit *unit)
{
    gdouble l2norm, q;

    l2norm = gwy_data_field_get_mean_square(dfield);

    /* In the integral formulation, we calculate the integral of squared
     * values and units of dx dy are reflected in the result.  In non-integral,
     * we calculate a mere sum of squared values and the result has the same
     * units as the field values. */
    if (as_integral) {
        q = gwy_data_field_get_xreal(dfield) * gwy_data_field_get_yreal(dfield);
        if (unit) {
            gwy_si_unit_multiply(gwy_data_field_get_si_unit_xy(dfield),
                                 gwy_data_field_get_si_unit_z(dfield),
                                 unit);
        }
    }
    else {
        q = gwy_data_field_get_xres(dfield) * gwy_data_field_get_yres(dfield);
        if (unit)
            gwy_si_unit_assign(unit, gwy_data_field_get_si_unit_z(dfield));
    }

    return sqrt(q*l2norm);
}

static void
preview(PSFControls *controls, PSFArgs *args)
{
    GwyLayerBasicRangeType range_type = GWY_LAYER_BASIC_RANGE_FULL;
    GwyDataField *dfield1, *dfield2, *psf, *convolved;
    GQuark quark;
    GwyContainer *data1, *data2;
    gdouble min, max, l2norm, resid;
    GwyResults *results;
    GwySIUnit *unit;
    guint i;

    data1 = gwy_app_data_browser_get(args->op1.datano);
    quark = gwy_app_get_data_key_for_id(args->op1.id);
    dfield1 = GWY_DATA_FIELD(gwy_container_get_object(data1, quark));

    if (args->op2.datano < 0 || args->op2.id < 0) {
        if (args->display == PSF_DISPLAY_DATA)
            gwy_data_field_copy(dfield1, controls->result, TRUE);
        else
            gwy_data_field_clear(controls->result);
        gwy_data_field_data_changed(controls->result);
        return;
    }

    data2 = gwy_app_data_browser_get(args->op2.datano);
    quark = gwy_app_get_data_key_for_id(args->op2.id);
    dfield2 = GWY_DATA_FIELD(gwy_container_get_object(data2, quark));

    psf = gwy_data_field_new_alike(dfield1, TRUE);

    calculate_tf(dfield1, dfield2, psf, args);

    convolved = gwy_data_field_duplicate(dfield2);
    gwy_data_field_add(convolved, -gwy_data_field_get_avg(convolved));
    field_convolve_default(convolved, psf);
    gwy_data_field_add(convolved, gwy_data_field_get_avg(dfield1));

    if (args->display == PSF_DISPLAY_DATA)
        gwy_data_field_assign(controls->result, dfield1);
    else if (args->display == PSF_DISPLAY_PSF)
        gwy_data_field_assign(controls->result, psf);
    else if (args->display == PSF_DISPLAY_CONVOLVED)
        gwy_data_field_assign(controls->result, convolved);
    else if (args->display == PSF_DISPLAY_DIFFERENCE) {
        gwy_data_field_assign(controls->result, convolved);
        gwy_data_field_subtract_fields(controls->result,
                                       dfield1, controls->result);
        range_type = GWY_LAYER_BASIC_RANGE_AUTO;
    }
    gwy_data_field_data_changed(controls->result);
    gwy_container_set_enum_by_name(controls->mydata, "/0/base/range-type",
                                   range_type);
    gwy_set_data_preview_size(GWY_DATA_VIEW(controls->view), PREVIEW_SIZE);
    /* Prevent the size changing wildly the moment someone touches the size
     * adjbars. */
    gtk_widget_set_size_request(controls->view, PREVIEW_SIZE, PREVIEW_SIZE);

    /* Change the normalisation to the discrete (i.e. wrong) one after all
     * calculations are done. */
    if (!args->as_integral)
        adjust_tf_to_non_integral(psf);

    gwy_data_field_subtract_fields(convolved, dfield1, convolved);
    results = controls->results;
    gwy_results_set_unit(results, "x", gwy_data_field_get_si_unit_xy(psf));
    gwy_results_set_unit(results, "y", gwy_data_field_get_si_unit_xy(psf));
    gwy_results_set_unit(results, "z", gwy_data_field_get_si_unit_z(psf));
    gwy_data_field_get_min_max(psf, &min, &max);
    unit = gwy_si_unit_new(NULL);
    l2norm = calculate_l2_norm(psf, args->as_integral, unit);
    gwy_results_set_unit(results, "u", unit);
    resid = calculate_l2_norm(convolved, args->as_integral, unit);
    gwy_results_set_unit(results, "v", unit);
    g_object_unref(unit);
    gwy_results_fill_values(results,
                           "width", measure_tf_width(psf),
                           "height", fmax(fabs(min), fabs(max)),
                           "l2norm", l2norm,
                           "residuum", resid,
                           NULL);
    for (i = 0; i < G_N_ELEMENTS(guivalues); i++) {
        gtk_label_set_markup(GTK_LABEL(controls->guivalues[i]),
                             gwy_results_get_full(results, guivalues[i]));
    }

    g_object_unref(psf);
    g_object_unref(convolved);
}

static void
psf_data_changed(GwyDataChooser *chooser, PSFControls *controls)
{
    GwyAppDataId *object = &controls->args->op2;

    gwy_data_chooser_get_active_id(chooser, object);
    update_sensitivity(controls);
    preview(controls, controls->args);
}

static gboolean
psf_data_filter(GwyContainer *data, gint id, gpointer user_data)
{

    GwyAppDataId *object = (GwyAppDataId*)user_data;
    GwyDataField *op1, *op2;
    GQuark quark;

    quark = gwy_app_get_data_key_for_id(id);
    op1 = GWY_DATA_FIELD(gwy_container_get_object(data, quark));

    data = gwy_app_data_browser_get(object->datano);
    quark = gwy_app_get_data_key_for_id(object->id);
    op2 = GWY_DATA_FIELD(gwy_container_get_object(data, quark));

    if (op1 == op2)
        return FALSE;

    return !gwy_data_field_check_compatibility(op1, op2,
                                               GWY_DATA_COMPATIBILITY_RES
                                               | GWY_DATA_COMPATIBILITY_REAL
                                               | GWY_DATA_COMPATIBILITY_LATERAL);
}

static void
calculate_tf(GwyDataField *measured, GwyDataField *ideal,
             GwyDataField *psf, PSFArgs *args)
{
    GwyDataField *wmeas, *wideal;
    gdouble r, sigma = pow10(args->sigma);
    gint xres, yres, xborder, yborder;

    wmeas = gwy_data_field_new_alike(measured, FALSE);
    wideal = gwy_data_field_new_alike(ideal, FALSE);
    prepare_field(measured, wmeas, args->windowing);
    prepare_field(ideal, wideal, args->windowing);
    if (args->method == PSF_METHOD_REGULARISED)
        gwy_data_field_deconvolve_regularized(wmeas, wideal, psf, sigma);
    else if (args->method == PSF_METHOD_PSEUDO_WIENER)
        psf_deconvolve_wiener(wmeas, wideal, psf, sigma);
    else {
        gwy_data_field_resample(psf, args->txres, args->tyres,
                                GWY_INTERPOLATION_NONE);
        gwy_data_field_deconvolve_psf_leastsq(wmeas, wideal, psf,
                                              sigma, args->border);
    }
    g_object_unref(wideal);
    g_object_unref(wmeas);

    if (!method_is_full_sized(args->method))
        return;

    xres = gwy_data_field_get_xres(psf);
    yres = gwy_data_field_get_yres(psf);
    xborder = (xres - args->txres + 1)/2;
    yborder = (yres - args->tyres + 1)/2;
    if (!xborder && !yborder)
        return;

    gwy_data_field_resize(psf,
                          xborder, yborder,
                          xborder + args->txres, yborder + args->tyres);
    r = (args->txres + 1 - args->txres % 2)/2.0;
    gwy_data_field_set_xoffset(psf, -gwy_data_field_jtor(psf, r));
    r = (args->tyres + 1 - args->tyres % 2)/2.0;
    gwy_data_field_set_yoffset(psf, -gwy_data_field_itor(psf, r));
}

static void
adjust_tf_to_non_integral(GwyDataField *psf)
{
    GwySIUnit *xyunit, *zunit;
    gdouble hxhy;

    xyunit = gwy_data_field_get_si_unit_xy(psf);
    zunit = gwy_data_field_get_si_unit_z(psf);
    gwy_si_unit_power_multiply(zunit, 1, xyunit, 2, zunit);

    hxhy = gwy_data_field_get_dx(psf) * gwy_data_field_get_dy(psf);
    gwy_data_field_multiply(psf, hxhy);
    gwy_data_field_data_changed(psf);
}

static gdouble
measure_tf_width(GwyDataField *psf)
{
    GwyDataField *mask, *abspsf;
    gint xres, yres;
    gdouble s2;

    xres = gwy_data_field_get_xres(psf);
    yres = gwy_data_field_get_yres(psf);
    mask = gwy_data_field_duplicate(psf);
    gwy_data_field_threshold(mask, 0.15*gwy_data_field_get_max(mask), 0.0, 1.0);
    if (gwy_data_field_get_val(mask, xres/2, yres/2) == 0.0) {
        g_object_unref(mask);
        return 0.0;
    }

    gwy_data_field_grains_extract_grain(mask, xres/2, yres/2);
    gwy_data_field_grains_grow(mask, 0.5*log(xres*yres),
                               GWY_DISTANCE_TRANSFORM_EUCLIDEAN, FALSE);
    abspsf = gwy_data_field_duplicate(psf);
    gwy_data_field_abs(abspsf);
    s2 = gwy_data_field_area_get_dispersion(abspsf, mask, GWY_MASK_INCLUDE,
                                            0, 0, xres, yres, NULL, NULL);
    g_object_unref(mask);
    g_object_unref(abspsf);

    return sqrt(s2);
}

static void
prepare_field(GwyDataField *field, GwyDataField *wfield,
              GwyWindowingType window)
{
    /* Prepare field in place if requested. */
    if (wfield != field) {
        gwy_data_field_resample(wfield,
                                gwy_data_field_get_xres(field),
                                gwy_data_field_get_yres(field),
                                GWY_INTERPOLATION_NONE);
        gwy_data_field_copy(field, wfield, TRUE);
    }
    gwy_data_field_add(wfield, -gwy_data_field_get_avg(wfield));
    gwy_fft_window_data_field(wfield, GWY_ORIENTATION_HORIZONTAL, window);
    gwy_fft_window_data_field(wfield, GWY_ORIENTATION_VERTICAL, window);
}

static gboolean
method_is_full_sized(PSFMethodType method)
{
    return (method == PSF_METHOD_REGULARISED
            || method == PSF_METHOD_PSEUDO_WIENER);
}

static void
estimate_tf_region(GwyDataField *wmeas, GwyDataField *wideal,
                   GwyDataField *psf,  /* scratch buffer */
                   gint *col, gint *row, gint *width, gint *height)
{
    gint xres, yres, i, j, imin, jmin, imax, jmax, ext;
    const gdouble *d;
    gdouble m;

    xres = gwy_data_field_get_xres(wmeas);
    yres = gwy_data_field_get_yres(wmeas);
    *col = xres/3;
    *row = yres/3;
    *width = xres - 2*(*col);
    *height = yres - 2*(*row);
    /* Use a fairly large but not yet insane sigma value 4.0 to estimate the
     * width.  We want to err on the side of size overestimation here.
     * XXX: We might want to use a proportional to 1/sqrt(xres*yres) here. */
    gwy_data_field_deconvolve_regularized(wmeas, wideal, psf, 4.0);
    d = gwy_data_field_get_data_const(psf);

    /* FIXME: From here it the same as to libprocess/filter.c
     * psf_sigmaopt_estimate_size(). */
    imax = yres/2;
    jmax = xres/2;
    m = 0.0;
    for (i = *row; i < *row + *height; i++) {
        for (j = *col; j < *col + *width; j++) {
            if (d[i*xres + j] > m) {
                m = d[i*xres + j];
                imax = i;
                jmax = j;
            }
        }
    }
    gwy_debug("maximum at (%d,%d)", imax, jmax);
    gwy_data_field_threshold(psf, 0.05*m, 0.0, 1.0);
    g_return_if_fail(d[imax*xres + jmax] > 0.0);
    gwy_data_field_grains_extract_grain(psf, jmax, imax);

    imin = imax;
    jmin = jmax;
    for (i = 0; i < yres; i++) {
        for (j = 0; j < xres; j++) {
            if (d[i*xres + j] > 0.0) {
                if (i < imin)
                    imin = i;
                if (i > imax)
                    imax = i;
                if (j < jmin)
                    jmin = j;
                if (j > jmax)
                    jmax = j;
            }
        }
    }

    ext = GWY_ROUND(0.5*log(xres*yres)) + 1;
    *col = jmin - ext;
    *row = imin - ext;
    *width = jmax+1 - jmin + 2*ext;
    *height = imax+1 - imin + 2*ext;
    if (*col < 0) {
        *width += *col;
        *col = 0;
    }
    if (*row < 0) {
        *height += *row;
        *row = 0;
    }
    if (*col + *width > xres)
        *width = xres - *col;
    if (*row + *height > yres)
        *height = yres - *row;

    gwy_debug("estimated region: %dx%d centered at (%d,%d)",
              *width, *height, *col + *width/2, *row + *height/2);

    /* Use some default reasonable size when things get out of hand... */
    *width = MIN(*width, xres/6);
    *height = MIN(*height, yres/6);
}

static void
symmetrise_tf_region(gint pos, gint len, gint res, gint *tres)
{
    gint epos = pos + len-1;
    len = MAX(epos, res-1 - pos) - MIN(pos, res-1 - epos) + 1;
    *tres = len | 1;
}

typedef struct {
    const PSFArgs *args;
    GwyDataField *psf;         /* Real-space PSF */
    GwyDataField *wideal;      /* Windowed ideal. */
    GwyDataField *wmeas;       /* Windowed measured. */
    gint col, row;
    gint width, height;
} PSFSigmaOptData;

static void
psf_sigmaopt_prepare(GwyDataField *measured, GwyDataField *ideal,
                     const PSFArgs *args,
                     PSFSigmaOptData *sodata)
{
    gwy_clear(sodata, 1);
    sodata->args = args;
    sodata->wideal = gwy_data_field_new_alike(ideal, FALSE);
    sodata->wmeas = gwy_data_field_new_alike(measured, FALSE);
    prepare_field(measured, sodata->wmeas, args->windowing);
    prepare_field(ideal, sodata->wideal, args->windowing);
    if (args->method == PSF_METHOD_PSEUDO_WIENER) {
        sodata->psf = gwy_data_field_new_alike(measured, FALSE);
        estimate_tf_region(sodata->wmeas, sodata->wideal, sodata->psf,
                           &sodata->col, &sodata->row,
                           &sodata->width, &sodata->height);
    }
}

static void
psf_sigmaopt_free(PSFSigmaOptData *sodata)
{
    GWY_OBJECT_UNREF(sodata->psf);
    g_object_unref(sodata->wmeas);
    g_object_unref(sodata->wideal);
}

static gdouble
psf_sigmaopt_evaluate(gdouble logsigma, gpointer user_data)
{
    PSFSigmaOptData *sodata = (PSFSigmaOptData*)user_data;
    const PSFArgs *args = sodata->args;
    GwyDataField *psf = sodata->psf;
    gdouble sigma, w;

    g_assert(args->method == PSF_METHOD_PSEUDO_WIENER);
    sigma = exp(logsigma);
    psf_deconvolve_wiener(sodata->wmeas, sodata->wideal, psf, sigma);
    gwy_data_field_area_abs(psf,
                            sodata->col, sodata->row,
                            sodata->width, sodata->height);
    w = gwy_data_field_area_get_dispersion(psf, NULL, GWY_MASK_IGNORE,
                                           sodata->col, sodata->row,
                                           sodata->width, sodata->height,
                                           NULL, NULL);
    return sqrt(w);
}

static gdouble
find_regularization_sigma(GwyDataField *dfield,
                          GwyDataField *ideal,
                          const PSFArgs *args)
{
    PSFSigmaOptData sodata;
    gdouble logsigma, sigma;

    g_return_val_if_fail(GWY_IS_DATA_FIELD(dfield), 0.0);
    g_return_val_if_fail(GWY_IS_DATA_FIELD(ideal), 0.0);
    g_return_val_if_fail(!gwy_data_field_check_compatibility
                                          (dfield, ideal,
                                           GWY_DATA_COMPATIBILITY_RES
                                           | GWY_DATA_COMPATIBILITY_REAL
                                           | GWY_DATA_COMPATIBILITY_LATERAL),
                         0.0);

    psf_sigmaopt_prepare(dfield, ideal, args, &sodata);
    if (args->method == PSF_METHOD_REGULARISED) {
        sigma = gwy_data_field_find_regularization_sigma_for_psf(sodata.wmeas,
                                                                 sodata.wideal);
    }
    else if (args->method == PSF_METHOD_LEAST_SQUARES) {
        sigma = gwy_data_field_find_regularization_sigma_leastsq(sodata.wmeas,
                                                                 sodata.wideal,
                                                                 args->txres,
                                                                 args->tyres,
                                                                 args->border);
    }
    else {
        logsigma = gwy_math_find_minimum_1d(psf_sigmaopt_evaluate,
                                            log(1e-8), log(1e3), &sodata);
        /* Experimentally determined fudge factor from large-scale
         * simulations. */
        sigma = 0.375*exp(logsigma);
    }
    psf_sigmaopt_free(&sodata);

    return sigma;
}

/* This is an exact replica of gwy_data_field_deconvolve_regularized().
 * The only difference is that instead of σ² the regularisation term is
 * σ²/|P|², corresponding to pseudo-Wiener filter with the assumption of
 * uncorrelated point noise. */
static void
psf_deconvolve_wiener(GwyDataField *dfield,
                      GwyDataField *operand,
                      GwyDataField *out,
                      gdouble sigma)
{
    gint xres, yres, i, cstride;
    gdouble lambda, q, frms, orms;
    fftw_complex *ffield, *foper;
    fftw_plan fplan, bplan;

    g_return_if_fail(GWY_IS_DATA_FIELD(dfield));
    g_return_if_fail(GWY_IS_DATA_FIELD(operand));
    g_return_if_fail(GWY_IS_DATA_FIELD(out));

    xres = dfield->xres;
    yres = dfield->yres;
    cstride = xres/2 + 1;
    g_return_if_fail(operand->xres == xres);
    g_return_if_fail(operand->yres == yres);
    gwy_data_field_resample(out, xres, yres, GWY_INTERPOLATION_NONE);

    orms = gwy_data_field_get_rms(operand);
    frms = gwy_data_field_get_rms(dfield);
    if (!orms) {
        g_warning("Deconvolution by zero.");
        gwy_data_field_clear(out);
        return;
    }
    if (!frms) {
        gwy_data_field_clear(out);
        return;
    }

    ffield = gwy_fftw_new_complex(cstride*yres);
    foper = gwy_fftw_new_complex(cstride*yres);
#if defined(_OPENMP) && defined(HAVE_FFTW_WITH_OPENMP)
    if (gwy_threads_are_enabled())
        fftw_plan_with_nthreads(gwy_omp_max_threads());
#endif
    fplan = fftw_plan_dft_r2c_2d(yres, xres, out->data, ffield,
                                 FFTW_DESTROY_INPUT);
    bplan = fftw_plan_dft_c2r_2d(yres, xres, ffield, out->data,
                                 FFTW_DESTROY_INPUT);

    gwy_data_field_copy(operand, out, FALSE);
    fftw_execute(fplan);
    gwy_assign(foper, ffield, cstride*yres);

    gwy_data_field_copy(dfield, out, FALSE);
    fftw_execute(fplan);
    fftw_destroy_plan(fplan);

    /* This seems wrong, but we just compensate the FFT. */
    orms *= sqrt(xres*yres);
    frms *= sqrt(xres*yres);
    lambda = sigma*sigma * orms*orms * frms*frms;
    /* NB: We normalize it as an integral.  So one recovers the convolution
     * with TRUE in ext-convolve! */
    q = 1.0/(dfield->xreal * dfield->yreal);
    for (i = 1; i < cstride*yres; i++) {
        gdouble fre = gwycreal(ffield[i]), fim = gwycimag(ffield[i]);
        gdouble ore = gwycreal(foper[i]), oim = gwycimag(foper[i]);
        gdouble inorm = ore*ore + oim*oim;
        gdouble fnorm = fre*fre + fim*fim;
        gdouble f = fnorm/(inorm*fnorm + lambda);
        gwycreal(ffield[i]) = (fre*ore + fim*oim)*f;
        gwycimag(ffield[i]) = (-fre*oim + fim*ore)*f;
    }
    fftw_free(foper);
    gwycreal(ffield[0]) = gwycimag(ffield[0]) = 0.0;
    fftw_execute(bplan);
    fftw_destroy_plan(bplan);
    fftw_free(ffield);

    gwy_data_field_multiply(out, q);
    gwy_data_field_2dfft_humanize(out);

    out->xreal = dfield->xreal;
    out->yreal = dfield->yreal;
    out->xoff = dfield->xoff;
    out->yoff = dfield->yoff;

    gwy_data_field_invalidate(out);
    set_transfer_function_units(operand, dfield, out);
}

static gboolean
clamp_psf_size(GwyDataField *dfield, PSFArgs *args)
{
    gint xres, yres;

    xres = gwy_data_field_get_xres(dfield);
    yres = gwy_data_field_get_yres(dfield);
    if (MIN(xres, yres) < 24)
        return FALSE;

    if (method_is_full_sized(args->method)) {
        args->txres = CLAMP(args->txres, 3, xres);
        args->tyres = CLAMP(args->tyres, 3, yres);
    }
    else {
        args->txres = CLAMP(args->txres, 3, xres/3 | 1);
        args->tyres = CLAMP(args->tyres, 3, yres/3 | 1);
    }
    args->border = CLAMP(args->border, 0, MIN(xres, yres)/8);
    return TRUE;
}

static const gchar as_integral_key[]  = "/module/psf/as_integral";
static const gchar border_key[]       = "/module/psf/border";
static const gchar display_key[]      = "/module/psf/display";
static const gchar method_key[]       = "/module/psf/method";
static const gchar output_type_key[]  = "/module/psf/output_type";
static const gchar sigma_key[]        = "/module/psf/sigma";
static const gchar txres_key[]        = "/module/psf/txres";
static const gchar tyres_key[]        = "/module/psf/tyres";
static const gchar windowing_key[]    = "/module/psf/windowing";

static void
psf_sanitize_args(PSFArgs *args)
{
    gwy_app_data_id_verify_channel(&args->op2);
    args->output_type &= PSF_OUTPUTS_MASK;
    args->sigma = CLAMP(args->sigma, -8.0, 3.0);
    args->as_integral = !!args->as_integral;
    args->display = CLAMP(args->display, 0, PSF_NDISPLAYS-1);
    args->method = CLAMP(args->method, 0, PSF_NMETHODS-1);
    args->windowing = gwy_enum_sanitize_value(args->windowing,
                                              GWY_TYPE_WINDOWING_TYPE);
}

static void
psf_load_args(GwyContainer *container, PSFArgs *args)
{
    *args = psf_defaults;

    gwy_container_gis_enum_by_name(container, method_key, &args->method);
    gwy_container_gis_double_by_name(container, sigma_key, &args->sigma);
    gwy_container_gis_enum_by_name(container, display_key, &args->display);
    gwy_container_gis_enum_by_name(container, output_type_key,
                                   &args->output_type);
    gwy_container_gis_boolean_by_name(container, as_integral_key,
                                      &args->as_integral);
    gwy_container_gis_enum_by_name(container, windowing_key, &args->windowing);
    gwy_container_gis_int32_by_name(container, txres_key, &args->txres);
    gwy_container_gis_int32_by_name(container, tyres_key, &args->tyres);
    gwy_container_gis_int32_by_name(container, border_key, &args->border);
    args->op2 = op2_id;

    psf_sanitize_args(args);
}

static void
psf_save_args(GwyContainer *container, PSFArgs *args)
{
    op2_id = args->op2;

    gwy_container_set_enum_by_name(container, method_key, args->method);
    gwy_container_set_double_by_name(container, sigma_key, args->sigma);
    gwy_container_set_enum_by_name(container, display_key, args->display);
    gwy_container_set_enum_by_name(container, output_type_key,
                                   args->output_type);
    gwy_container_set_boolean_by_name(container, as_integral_key,
                                      args->as_integral);
    gwy_container_set_enum_by_name(container, windowing_key, args->windowing);
    gwy_container_set_int32_by_name(container, txres_key, args->txres);
    gwy_container_set_int32_by_name(container, tyres_key, args->tyres);
    gwy_container_set_int32_by_name(container, border_key, args->border);
}

/* vim: set cin et ts=4 sw=4 cino=>1s,e0,n0,f0,{0,}0,^0,\:1s,=0,g1s,h0,t0,+1s,c3,(0,u0 : */
